// *****************************************************************************
// * PROJECT:  Mouse routines with 'real' graphic cursor in text mode.
// *****************************************************************************
// * MODULE:  MOU.H
// *****************************************************************************
// * DESCRIPTION:
// *   Header file for the mouse routines.
// *
// *****************************************************************************
// * MODIFICATION NOTES:
// *    Date     Author Comment
// * 26-Oct-1990   dk   Initial file.
// * 07-Jan-1991   dk   Fixed bugs and set up for release to Usenet.
// *****************************************************************************
// *
// * DISCLAIMER:
// *
// * Programmers may incorporate any or all code into their programs,
// * giving proper credit within the source. Publication of the
// * source routines is permitted so long as proper credit is given
// * to Dave Kirsch.
// *
// * Copyright (C) 1990, 1991 by Dave Kirsch.  You may use this program, or
// * code or tables extracted from it, as desired without restriction.
// * I can not and will not be held responsible for any damage caused from
// * the use of this software.
// *
// *****************************************************************************
// * This source works with Turbo C 2.0 and MSC 6.0 and above.
// *****************************************************************************

// ********************************************************
// * 26-Oct-1990 - dk                                     *
// *                                                      *
// *  Standard types and constants I use in my programs.  *
// *                                                      *
// ********************************************************

#include <stdio.h>
#include <dos.h>
#include <stdlib.h>
#include <string.h>

typedef unsigned char       byte;
typedef unsigned short int  word;
typedef unsigned short int  boolean;
typedef unsigned long  int  dword;

#define FALSE 0
#define TRUE (!FALSE)

#ifdef __TURBOC__
  #define FAST pascal     // for fast calling of functions -- Turbo C 
#else
  #define FAST _fastcall  // for fast calling of functions -- MicroSoft C 6.0 
  #define asm _asm
#endif
#define LOCAL   near   // function can not be called outside of this module 
#define PRIVATE static // function is private 
#define STATIC  static // private variables 

#ifndef poke
#define poke(a,b,c)     (*((int  far*)MK_FP((a),(b))) = (int)(c))
#define pokeb(a,b,c)    (*((char far*)MK_FP((a),(b))) = (char)(c))
#define peek(a,b)       (*((int  far*)MK_FP((a),(b))))
#define peekb(a,b)      (*((char far*)MK_FP((a),(b))))
#endif

// ***************************************************
// * Mon 07-Jan-1991 - dk                            *
// *                                                 *
// *  Variables and defines for the mouse routines.  *
// *                                                 *
// ***************************************************

// Size of mouse "click" ahead buffer.
#define MOUSEBUFFERSIZE 16

// Bit defines for mouse driver function 12 -- define handler. 
#define MOUSEMOVE      1
#define LEFTBPRESS     2
#define LEFTBRELEASE   4
#define RIGHTBPRESS    8
#define RIGHTBRELEASE 16

#define LEFTBDOWN  1
#define RIGHTBDOWN 2

// Shift states for byte a 0:417h
//   bit 7 =1 INSert active
//   bit 6 =1 Caps Lock active
//   bit 5 =1 Num Lock active
//   bit 4 =1 Scroll Lock active
//   bit 3 =1 either Alt pressed
//   bit 2 =1 either Ctrl pressed
//   bit 1 =1 Left Shift pressed
//   bit 0 =1 Right Shift pressed
//

#define SHIFT_RIGHTSHIFT 0x01
#define SHIFT_LEFTSHIFT  0x02
#define SHIFT_SHIFT      0x03 // Either shift key. 
#define SHIFT_CTRL       0x04
#define SHIFT_ALT        0x08
#define SHIFT_SCROLLLOCK 0x10
#define SHIFT_NUMLOCK    0x20
#define SHIFT_CAPSLOCK   0x40
#define SHIFT_INS        0x80

// Mouse information record 
struct minforectype {
  word buttonstat;
  int  cx, cy;
  byte shiftstate;
};

#define MOUINFOREC struct minforectype

#ifdef __TURBOC__
#pragma inline
#endif

word mousehidden = 0;           // Is the mouse on? Additive flag 
boolean mouseinstalled = FALSE; // Is the mouse installed? 

volatile int mousex, mousey; // Character position of mouse 

#ifndef __TURBOC__
	STATIC word _based(_segname("_CODE")) DGroupSeg;
#endif

STATIC volatile int mbufin = 0, mbufout = 0; // Mouse buffer pointers 

STATIC word mousefreeze = 0;           // Is mouse frozen in place? 

STATIC MOUINFOREC mbuf[MOUSEBUFFERSIZE]; // Mouse buffer 

STATIC word oldword;
STATIC word newword;
STATIC boolean saved = FALSE;
STATIC word oldmx, oldmy;


STATIC boolean egavga = FALSE; 
STATIC byte savechars[3][3]; 
STATIC byte chardefs[32 * 9]; 
STATIC word mousepx, mousepy; 

STATIC boolean conditionalhidemouse = FALSE;
STATIC word conx1, cony1, conx2, cony2;

STATIC word vseg; // Segment of video ram.
STATIC word mcols, mrows;
STATIC byte savevmode;

STATIC int maxx, maxy;

STATIC boolean desqview = FALSE;

#define POKEATTRIB(x, y, a) pokeb(vseg, (y) * (mcols * 2) + ((x) << 1) + 1, a)
#define PEEKATTRIB(x, y)    peekb(vseg, (y) * (mcols * 2) + ((x) << 1) + 1)

#define POINTS *((byte far *) 0x00000485)
#define COLS *((byte far *) 0x0040004AL)
#define ROWS *((byte far *) 0x00400084L)

#define DEFCHAR 0xd0


PRIVATE void far mousehandler(void)
//  This function is called whenever a button is pressed.  Do not call this
//   function directly!!

{
	register int conditionmask;
	asm push ds

	#ifdef __TURBOC__
		asm push ax
		asm mov ax, DGROUP
		asm mov ds, ax
		asm pop ax
	#else
		asm mov ds, cs:DGroupSeg
	#endif

	asm mov conditionmask,ax

	if (!mousefreeze) 
	{
		asm mov mousex, cx
		asm mov mousey, dx
		asm mov mousepx, cx
		asm mov mousepy, dx

		oldmx = mousex;
		oldmy = mousey;
		saved = TRUE;
	}
	// Now, see if a mouse button was whacked 
	if (conditionmask & ~MOUSEMOVE)
	{
		mbuf[mbufin].buttonstat = conditionmask & ~MOUSEMOVE;
		mbuf[mbufin].cx = mousex;
		mbuf[mbufin].cy = mousey;
		mbuf[mbufin].shiftstate = peekb(0, 0x417); // Get shift byte 
		mbufin = (mbufin + 1) % MOUSEBUFFERSIZE;
	}
	asm pop ds;
}

// ************************************
// * 26-Oct-1990 - dk                 *
// *                                  *
// *  Initialize the mouse routines.  *
// *                                  *
// ************************************
void FAST MOUinit(void)
{
	byte v;
	word points;

	#ifndef __TURBOC__
		asm mov cs:DGroupSeg,ds   // save for interrupt handler to use
	#endif

	asm sub ax,ax;    // Mouse driver function 0 -- reset and detect 
	asm int 33h
	asm mov mouseinstalled, AX;

	if (mouseinstalled)
	{ 
		mousefreeze++; // Make sure handler doesn't do things, yet

		savevmode = peekb(0x40, 0x49);
		pokeb(0x40, 0x49, 111); // Does this work ?!?!?!?!? 

		// Reset driver for change in video mode to take effect.
		asm sub ax,ax
		asm int 33h

		//Now that we've tricked the mouse driver into a grapics mode thereby
		//causing it to give us pixel resolutions, put the old mode back. 
		pokeb(0x40, 0x49, savevmode);

		// Set up max x and y ranges.

		maxx = 800 - 1;
		maxy = 600 - 1; 

		asm mov dx,maxx     // Pixels horizontally 
		asm mov ax,7        // mouse driver function 7 -- set max x range 
		asm sub cx,cx       // Minimum range 
		asm int 33h

		asm mov dx,maxy     // Pixels veritcally 
		asm mov ax,8        // mouse driver function 8 -- set max y range 
		asm sub cx,cx       // Minimum range 
		asm int 33h

		// Now install user routine 

		asm mov ax,cs
		asm mov es,ax
		asm mov dx, offset mousehandler
		// Setup up bits for calling routine 
		#ifdef __TURBOC__
			_CX = LEFTBPRESS | LEFTBRELEASE | RIGHTBPRESS | RIGHTBRELEASE | MOUSEMOVE;
		#else
			asm mov cx,LEFTBPRESS | LEFTBRELEASE | RIGHTBPRESS | RIGHTBRELEASE | MOUSEMOVE;
		#endif
		asm mov ax,12       // Function 12 -- set user routine 
		asm int 33h

		mousex = mousey = 0;
		asm mov cx,mousex   // xcoord 
		asm mov dx,mousey   // ycoord 
		asm mov ax,4    // mouse driver function 4 -- set mouse position 
		asm int 33h

		mousefreeze--; // Handler can get into business, now 
	}
}


// *************************************************************
// * 27-Oct-1990 - dk                                          *
// *                                                           *
// *  Returns true if there is something in the mouse buffer.  *
// *                                                           *
// *************************************************************
boolean FAST MOUcheck(void)
{
  return mbufin != mbufout;
}

// **************************************************************
// * 26-Oct-1990 - dk                                           *
// *                                                            *
// *  Take a copy of the mouse event at the head of the queue.  *
// *                                                            *
// **************************************************************
void FAST MOUpreview(MOUINFOREC *mouinforec)
{
  if (!mouseinstalled)
    return;

  if (mbufin != mbufout) 
    *mouinforec = mbuf[mbufout];
  else
  {
    mouinforec -> cx = mousex;
    mouinforec -> cy = mousey;
    mouinforec -> buttonstat = 0;
    mouinforec -> shiftstate = peekb(0, 0x417);
  }
}

// ****************************************************************
// * 26-Oct-1990 - dk                                             *
// *                                                              *
// *  Get (and remove) the mouse event at the head of the queue.  *
// *                                                              *
// ****************************************************************
void FAST MOUget(MOUINFOREC *mouinforec)
{
  if (!mouseinstalled)
    return;

  if (mbufin != mbufout)
  { 
    if (mouinforec != NULL)
      *mouinforec = mbuf[mbufout];
    mbufout = (mbufout + 1) % MOUSEBUFFERSIZE;
  } 
  else
  {
    mouinforec -> cx = mousex;
    mouinforec -> cy = mousey;
    mouinforec -> buttonstat = 0;
    mouinforec -> shiftstate = peekb(0, 0x417);
  }
}

// **************************************
// * 26-Oct-1990 - dk                   *
// *                                    *
// *  Deinitialize the mouse routines.  *
// *                                    *
// **************************************
void FAST MOUdeinit(void)
{
  if (!mouseinstalled)
    return;

  asm xor ax,ax
  asm int 33h
}

// **************************************************
// * 26-Oct-1990 - dk                               *
// *                                                *
// *  Returns the bits for the button status info.  *
// *                                                *
// **************************************************
word FAST MOUbuttonstatus(void)
{
word buts;

  if (!mouseinstalled)
    return 0;

  asm mov ax,3
  asm int 33h
  asm mov buts,bx
  return buts;
}
